"""License:
Distributed under the OSI-approved Apache License, Version 2.0.  See
accompanying file Copyright.txt for details.
"""

from functools import singledispatchmethod
from sys import maxsize
from adios2 import bindings, Adios, Stream, IO


# pylint: disable=W0221
# pylint: disable=R0902   # Too many instance attributes
class FileReader(Stream):
    """High level implementation of the FileReader class for read Random access mode"""

    def __repr__(self):
        return f"<adios.file named {self._io_name}>"

    @singledispatchmethod
    def __init__(self, path: str, comm=None):
        super().__init__(path, "rra", comm)

    # e.g. FileReader(io: adios2.IO, path, mode)
    # pylint: disable=E1121
    @__init__.register(IO)
    def _(self, io: IO, path: str, comm=None):
        super().__init__(io, path, "rra", comm)

    # e.g. FileReader(path, metadata)
    # pylint: disable=E1121
    @__init__.register(bytes)
    def _(self, metadata: bytes, path: str):
        self._adios = Adios()
        self._io_name = f"stream:{path}:mode:rra"
        # pylint: enable=E1121
        self._io = self._adios.declare_io(self._io_name)
        self._mode = bindings.Mode.ReadRandomAccess
        self._read_mode = True
        self._engine = self._io.open_with_metadata(path, metadata)
        self.index = -1
        self.max_steps = maxsize
        self._step_status = bindings.StepStatus.EndOfStream
        self._step_timeout_sec = self.DEFAULT_TIMEOUT_SEC

    # pylint: enable=E1121



from os import fstat

import numpy as np


def ReadMetadataStep(f, fileSize, MetadataEntry, WriterMapEntry):
    # Read metadata of one step
    step = MetadataEntry['step']
    mdpos = MetadataEntry['mdpos']
    mdsize = MetadataEntry['mdsize']
    # flushcount = MetadataEntry['flushcount']
    WriterCount = WriterMapEntry['WriterCount']

    if mdpos + mdsize > fileSize:
        print(f"ERROR: step {step} metadata pos {mdpos} + size {mdsize} "
              f"is beyond the metadata file size {fileSize}")
        return False

    currentpos = f.tell()
    if mdpos > currentpos:
        print(f"Offset {currentpos}..{mdpos-1} is a gap unaccounted for")

    if mdpos < currentpos:
        print(f"ERROR: step {step} metadata pos {mdpos} points before the "
              f"expected position in file {currentpos}")
        return False

    f.seek(mdpos)
    buf = f.read(mdsize)
    pos = 0

    if step > 0:
        print("========================================================")
    print(f"Step {step}: ")
    print(f"  Offset = {mdpos}")

    mdsize_in_file = np.frombuffer(buf, dtype=np.uint64, count=1,
                                   offset=pos)
    pos = pos + 8
    if (mdsize == mdsize_in_file[0] + 8):
        print(f"  Size = {mdsize_in_file[0]}")
    else:
        print(f"ERROR: md record supposed to be {mdsize-8} + 8 bytes "
              f"(as recorded in index), but found in file "
              f"{mdsize_in_file[0]}")

    MDPosition = mdpos + 2 * 8 * WriterCount
    print("  Variable metadata entries: ")
    for w in range(0, WriterCount):
        a = np.frombuffer(buf, dtype=np.uint64, count=1, offset=pos)
        thisMDSize = int(a[0])
        pos = pos + 8
        print(f"    Writer {w}: md size {thisMDSize} "
              f"offset {MDPosition}")
        MDPosition = MDPosition + thisMDSize

    print("  Attribute metadata entries: ")
    for w in range(0, WriterCount):
        a = np.frombuffer(buf, dtype=np.uint64, count=1, offset=pos)
        thisMDSize = int(a[0])
        pos = pos + 8
        print(f"    Writer {w}: md size {thisMDSize} "
              f"offset {MDPosition}")
        MDPosition = MDPosition + thisMDSize

    if (mdsize_in_file != MDPosition - mdpos):
        print(f"ERROR: entries supposed to end at start offset+size "
              f"{mdpos}+{mdsize_in_file[0]}, but it ends instead on offset "
              f"{MDPosition}")

    return True


def DumpMetaData(fileName, MetadataIndexTable, WriterMap):
    print("========================================================")
    print("    Metadata File: " + fileName)
    print("========================================================")

    # print(f"MetadataIndexTable={MetadataIndexTable}")
    # print(f"WriterMap={WriterMap}")

    with open(fileName, "rb") as f:
        fileSize = fstat(f.fileno()).st_size
        for MetadataEntry in MetadataIndexTable:
            WriterMapEntry = WriterMap[MetadataEntry['writermapindex']]
            status = ReadMetadataStep(
                f, fileSize, MetadataEntry, WriterMapEntry)
    return status


if __name__ == "__main__":
    print("ERROR: Utility main program is bp5dbg.py")



from adios2 import *
import adios2.bindings as bindings

import sys
import unittest


DATA_FILENAME = "hello-world-py.bp"

class TestSimpleReadWrite(unittest.TestCase):

    def _write(self, ad, greeting):
        """write a string to a bp file"""
        io = ad.declare_io("hello-world-writer")
        var_greeting = io.define_variable("Greeting")
        w = io.open(DATA_FILENAME, bindings.Mode.Write)
        w.begin_step()
        w.put(var_greeting, greeting)
        w.end_step()
        w.close()
        return 0

    def _read(self, ad):
        """read a string from to a bp file"""
        io = ad.declare_io("hello-world-reader")
        r = io.open(DATA_FILENAME, bindings.Mode.Read)
        r.begin_step()
        var_greeting = io.inquire_variable("Greeting")
        message = r.get(var_greeting)
        r.end_step()
        r.close()
        return message

    def test_simple_read_write(self):
        """driver function"""
        print("ADIOS2 version {0}".format(adios2.__version__))
        ad = adios2.Adios()
        greeting = "Hello World from ADIOS2"
        self._write(ad, greeting)
        message = self._read(ad)
        print("{}".format(message))
        self.assertEqual(greeting, message)
        return 0


if __name__ == '__main__':
    unittest.main()



"""License:
Distributed under the OSI-approved Apache License, Version 2.0.  See
accompanying file Copyright.txt for details.
"""


class Operator:
    """High level representation of the Attribute class in the adios2.bindings"""

    def __init__(self, implementation, name):
        self.impl = implementation
        self._name = name

    @property
    def impl(self):
        """Bindings implementation of the class"""
        return self._impl

    @impl.setter
    def impl(self, implementation):
        self._impl = implementation

    def __eq__(self, other):
        if isinstance(other, Operator):
            return self._name == other._name
        return False

    def get_parameters(self):
        """
        Get parameters associated to this Operator

        Returns:
            dict: parameters
        """
        return self.impl.Parameters()

    def set_parameter(self, key, value):
        """
        Set parameter associated to this Operator

        Args:
            str: key
            str: value
        """
        self.impl.SetParameter(key, value)



"""License:
Distributed under the OSI-approved Apache License, Version 2.0.  See
accompanying file Copyright.txt for details.
"""

import numpy as np
from adios2.attribute import Attribute
from adios2.variable import Variable
from adios2.derived_variable import DerivedVariable
from adios2.engine import Engine


class IO:
    """High level representation of the IO class in the adios2.bindings"""

    def __init__(self, impl, name, adiosobj):
        self.impl = impl
        self._name = name
        self._adios = adiosobj

    @property
    def impl(self):
        """Bindings implementation of the class"""
        return self._impl

    @impl.setter
    def impl(self, implementation):
        self._impl = implementation

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.flush_all()

    def __eq__(self, other):
        if isinstance(other, IO):
            return self._name == other._name
        return False

    def adios(self):
        """Adios instance associated to this IO"""
        return self._adios

    def define_attribute(
        self,
        name,
        content=None,
        variable_name="",
        separator="/",
    ):
        """
        Define an Attribute

        Parameters
            name
                attribute name

            content
                attribute numpy array data

            variable_name
                if attribute is associated with a variable

            separator
                concatenation string between variable_name and attribute
                e.g. variable_name + separator + name ("var/attr")
                Not used if variable_name is empty
        """

        if isinstance(content, (np.ndarray, str, list)):
            return Attribute(self.impl, name, content, variable_name, separator)

        if not hasattr(content, "__len__") and isinstance(content, (int, float, complex)):
            return Attribute(self.impl, name, content, variable_name, separator)

        # Anything else, try to pass as a numpy array (and get an error here if that fails)
        return Attribute(self.impl, name, np.asarray(content), variable_name, separator)

    def inquire_attribute(self, name, variable_name="", separator="/"):
        """
        Inquire an Attribute

        Parameters
            name
                attribute name

            variable_name
                if attribute is associated with a variable

            separator
                concatenation string between variable_name and attribute
                e.g. variable_name + separator + name ("var/attr")
                Not used if variable_name is empty
        """
        attr = None
        attrimpl = self.impl.InquireAttribute(name, variable_name, separator)
        if attrimpl:
            attr = Attribute.__new__(Attribute)
            attr.impl = attrimpl
        return attr

    def available_attributes(self, varname="", separator="/"):
        """
        Returns a 2-level dictionary with attribute information.
        Read mode only.

        Parameters
            variable_name
                If varname is set, attributes assigned to that variable are returned.
                The keys returned are attribute names with the prefix of varname + separator
                removed.

            separator
                concatenation string between variable_name and attribute
                e.g. varname + separator + name ("var/attr")
                Not used if varname is empty

        Returns
            attributes dictionary
                key
                    attribute name
                value
                    attribute information dictionary
        """
        return self.impl.AvailableAttributes(varname, separator)

    def remove_attribute(self, name):
        """
        Remove an Attribute

        Parameters
            name
                attribute name
        """
        self.impl.RemoveAttribute(name)

    def remove_all_attributes(self):
        """
        Remove all define attributes
        """
        self.impl.RemoveAllAttributes()

    def define_variable(
        self,
        name,
        content=None,
        shape=[],
        start=[],
        count=[],
        is_constant_dims=False,
    ):
        """
        writes a variable

        Parameters
            name
                variable name

            content
                variable data values

            shape
                variable global MPI dimensions.

            start
                variable offset for current MPI rank.

            count
                variable dimension for current MPI rank.

            isConstantDims
                Whether dimensions are constant
        """
        var_impl = None
        if isinstance(content, list) and len(content) > 0:
            var_impl = self.impl.DefineVariable(
                name, content[0], shape, start, count, is_constant_dims
            )

        elif content is None:
            var_impl = self.impl.DefineVariable(name, "", shape, start, count, is_constant_dims)

        else:
            var_impl = self.impl.DefineVariable(
                name, content, shape, start, count, is_constant_dims
            )

        return Variable(var_impl)

    def inquire_variable(self, name):
        """
        Inquire a variable

        Parameters
            name
                variable name
        Returns
            The variable if it is defined, otherwise None
        """
        var = None
        var_impl = self.impl.InquireVariable(name)
        if var_impl:
            var = Variable.__new__(Variable)
            var.impl = var_impl
        return var

    def available_variables(self):
        """

        Returns a 2-level dictionary with variable information.
        Read mode only.

        Parameters
            keys
               list of variable information keys to be extracted (case insensitive)
               keys=['AvailableStepsCount','Type','Max','Min','SingleValue','Shape']
               keys=['Name'] returns only the variable names as 1st-level keys
               leave empty to return all possible keys

        Returns
            variables dictionary
                key
                    variable name
                value
                    variable information dictionary
        """
        return self.impl.AvailableVariables()

    def remove_variable(self, name):
        """
        Remove a variable

        Parameters
            name
                Variable name
        """
        self.impl.RemoveVariable(name)

    def remove_all_variables(self):
        """
        Remove all variables in the IO instance
        """
        self.impl.RemoveAllVariables()

    def define_derived_variable(self, name, expression, etype=None):
        """
        Define a derived variable with an expression

        Parameters
            name
                name as it appears in variable list in the output

            expression
                expression string using other variable names, operators and functions

            type
                DerivedVarType.StatsOnly     : store only the metadata of the derived variable
                DerivedVarType.ExpressionString : store only the definition, nothing else
                DerivedVarType.StoreData        : store as a complete variable (data and metadata)
        """
        var_impl = None

        if etype is None:
            var_impl = self.impl.DefineDerivedVariable(name, expression)
        else:
            var_impl = self.impl.DefineDerivedVariable(name, expression, etype)

        return DerivedVariable(var_impl)

    def open(self, name, mode, comm=None):
        """
        Open an engine

        Parameters
            name
                Engine name

            mode
                engine mode

            comm
                MPI communicator, optional
        """

        if comm:
            return Engine(self.impl.Open(name, mode, comm))

        return Engine(self.impl.Open(name, mode))

    def open_with_metadata(self, name, metadata: bytes):
        """
        Open an engine with metadata already in memory

        Parameters
            name
                Engine name

            metadata
                file metadata retrieved by FileReader.get_metadata()
        """

        return Engine(self.impl.Open(name, metadata))

    def set_engine(self, name):
        """
        Set engine for this IO instance

        Args:
            name (str): name of engine
        """
        self.impl.SetEngine(name)

    def engine_type(self):
        """Return engine type"""
        return self.impl.EngineType()

    def add_transport(self, kind, parameters={}):
        """
        Adds a transport and its parameters to current IO. Must be
        supported by current engine type.

        Parameters
            kind
                must be a supported transport type for current engine.

            parameters
                acceptable parameters for a particular transport
                CAN'T use the keywords "Transport" or "transport" in key

        Returns
            transport_index
                handler to added transport
        """
        return self.impl.AddTransport(kind, parameters)

    def parameters(self):
        """
        Return parameter associated to this io instance

        Return:
            dict: str->str parameters
        """
        return self.impl.Parameters()

    def set_parameter(self, key, value):
        """
        Set a parameter for this IO instance

        Args:
            key (str):
            value (str):
        """
        self.impl.SetParameter(key, value)

    def set_parameters(self, parameters):
        """
        Set parameters for this IO instance

        Args:
            parameters (dict):
        """
        self.impl.SetParameters(parameters)

    def flush_all(self):
        """Flush all engines attached to this IO instance"""
        self.impl.FlushAll()



"""License:
Distributed under the OSI-approved Apache License, Version 2.0.  See
accompanying file Copyright.txt for details.
"""

from adios2 import bindings


class Variable:
    """High level representation of the Variable class in the adios2.bindings"""

    def __init__(self, implementation):
        self.impl = implementation

    @property
    def impl(self):
        """Bindings implementation of the class"""
        return self._impl

    @impl.setter
    def impl(self, implementation):
        self._impl = implementation

    def __eq__(self, other):
        if isinstance(other, Variable):
            return self.name() == other.name()
        return False

    def block_id(self):
        """
        BlockID of this variable.

        Returns:
            int: BlockID of this variable.
        """
        return self.impl.BlockID()

    def count(self):
        """
        Current selected count for this variable.

        Returns:
            int: Current selected count.
        """
        return self.impl.Count()

    def selection_size(self):
        """
        Current selection size selected for this variable.

        Returns:
            int: Current selection size selected.
        """
        return self.impl.SelectionSize()

    def set_block_selection(self, block_id):
        """
        Set BlockID for this variable.

        Args:
            block_id (int): Selected BlockID.
        """
        self.impl.SetBlockSelection(block_id)

    def set_selection(self, selection):
        """
        Set selection for this variable.

        Args:
            selection (list): list of the shape [[start], [count]], note that start and
            count can contain more than one element.
        """
        self.impl.SetSelection(selection)

    def set_shape(self, shape):
        """
        Set Shape (dimensions) for this variable

        Args:
            shape (list): desired shape (dimensions).
        """
        self.impl.SetShape(shape)

    def set_step_selection(self, step_selection):
        """
        Set current step selection (For RRA or ReadRandomAccess)

        Args:
            step_selection (list): On the form of [start, count].
        """
        self.impl.SetStepSelection(step_selection)

    def set_accuracy(self, error, norm, relative):
        """
        Set Accuracy for (remote) reading for this variable

        Args:
            error: floating point value
            norm:  floating point value
            relative: True or False
        """
        acc = bindings.Accuracy(error, norm, relative)
        self.impl.SetAccuracy(acc)

    def shape(self, step=None):
        """
        Get the shape assigned to the given step for this variable.

        Args:
            step (int): Desired step. Only in ReadRandomAccess mode

        Returns:
            list: shape of the specified step in the form of [start, count].
        """
        if step is None:
            return self.impl.Shape()
        return self.impl.Shape(step)

    def shape_id(self):
        """
        Get the ShapeID assigned to this variable.

        Returns:
            int: ShapeID assigned to this variable.
        """
        return self.impl.ShapeID()

    def type(self):
        """
        Type of the Variable

        Returns:
            str: Type of the Variable.
        """
        return self.impl.Type()

    def single_value(self):
        """
        Check if this variable is a single value.

        Returns:
            bool: true if this variable is a single value.
        """
        return bool(self.impl.SingleValue())

    def sizeof(self):
        """
        Size in bytes of the contents of the variable.

        Returns:
            int: size in bytes of the contents.
        """
        return self.impl.Sizeof()

    def start(self):
        """
        The current selected start of the variable.

        Returns:
            int: The current selected start of the variable.
        """
        return self.impl.Start()

    def steps(self):
        """
        The number of steps of the variable. This is always 1 in a stream.
        In ReadRandomAccess mode, this function returns the total number
        of steps available, which can be used when selecting steps for read.

        Returns:
            int: The avaialble steps of the variable.
        """
        return self.impl.Steps()

    def steps_start(self):
        """
        The avaliable start step, this is needed variables can start
        at any time step. This is for ReadRandomAccess.

        Returns:
            int: the starting step of for this Variable.
        """
        return self.impl.StepsStart()

    def name(self):
        """
        Name of the Variable

        Returns:
            str: Name of the Variable.
        """
        return self.impl.Name()

    def get_accuracy(self):
        """
        Get the accuracy of the variable (of its last read)

        Returns:
            adios2.bindings.Accuracy struct (with error, norm and relative fields).
        """
        return self.impl.GetAccuracy()

    def add_operation_string(self, name, params={}):
        """
        Add an operation (operator) as a string

        Args:
            name (str): name of the operation.
            params (dict): parameters as a form of a dict for the operation.
        """
        return self.impl.AddOperation(name, params)

    def add_operation(self, operation, params={}):
        """
        Add an operation (operator).

        Args:
            name (Operator): name of the operation.
            params (dict): parameters as a form of a dict for the operation.
        """
        return self.impl.AddOperation(operation.impl, params)

    def operations(self):
        """
        Current operations (operators) assigned to this Variable.

        Returns:
            list(Operators): operators assigned.
        """
        return self.impl.Operations()

    def remove_operations(self):
        """
        Remove operations (operators) assigned to this Variable.
        """
        self.impl.RemoveOperations()

    def store_stats_only(self, mode):
        """
        Select writing mode for variable (write data or just stats)
        """
        self.impl.StoreStatsOnly(mode)




LocalValueDim = 18446744073709551613

dataTypes = {
    -1: 'unknown',
    0: 'byte',
    1: 'short',
    2: 'integer',
    4: 'long',

    50: 'unsigned_byte',
    51: 'unsigned_short',
    52: 'unsigned_integer',
    54: 'unsigned_long',

    5: 'real',
    6: 'double',
    7: 'long_double',

    9: 'string',
    10: 'complex',
    11: 'double_complex',
    12: 'string_array',

    55: 'char'
}

dataTypeSize = {
    -1: 0,
    0: 1,
    1: 2,
    2: 4,
    4: 8,

    50: 1,
    51: 2,
    52: 4,
    54: 8,

    5: 4,
    6: 8,
    7: 16,

    9: 0,
    10: 8,
    11: 16,
    12: 0,

    55: 1
}


def GetTypeName(typeID):
    name = dataTypes.get(typeID)
    if name is None:
        name = "unknown type"
    return name


def GetTypeSize(typeID):
    size = dataTypeSize.get(typeID)
    if size is None:
        size = 0
    return size


CharacteristicNames = {
    0: 'value',
    1: 'min',
    2: 'max',
    3: 'offset',
    4: 'dimensions',
    5: 'var_id',
    6: 'payload_offset',
    7: 'file_index',
    8: 'time_index',
    9: 'bitmap',
    10: 'stat',
    11: 'transform_type',
    12: 'minmax'
}


def GetCharacteristicName(cID):
    name = CharacteristicNames.get(cID)
    if name is None:
        name = "unknown characteristic"
    return name


def GetCharacteristicDataLength(cID, typeID):
    name = CharacteristicNames.get(cID)
    if (name == 'value' or name == 'min' or
            name == 'max' or name == 'minmax'):
        return dataTypeSize[typeID]
    elif (name == 'offset' or name == 'payload_offset'):
        return 8
    elif (name == 'file_index' or name == 'time_index'):
        return 4
    else:
        return 0


# Read Header info 64 bytes
# fileType: Data, Metadata, Index Table
def ReadHeader(f, fileSize, fileType, verbose):
    status = True
    if fileSize < 64:
        print("ERROR: Invalid " + fileType + ". File is smaller "
              "than the header (64 bytes)")
        return False
    header = f.read(64)
    hStr = header.decode('ascii')

    versionStr = hStr[0:32].replace('\0', ' ')
    major = hStr[32]
    minor = hStr[33]
    micro = hStr[34]
#    unused = hStr[35]

    endianValue = header[36]
    if endianValue == 0:
        endian = 'yes'
    elif endianValue == 1:
        endian = ' no'
    else:
        print("ERROR: byte 28 must be 0 or 1 to indicate endianness of "
              "the data. It is however {0} in this file".format(
                  endianValue))
        status = False

    bpversion = int(header[37])
    bpminorversion = int(header[38])
    active = int(header[39])
    if active == 0:
        activeStr = ' no'
    else:
        activeStr = 'yes'
    iscolumnmajor = hStr[40]
    if iscolumnmajor == 'n':
        clmnStr = '   no'
    else:
        clmnStr = '  yes'
    # 45..63 unused

    if not verbose:
        return status

    print("---------------------------------------------------------"
          "---------------------------------------------------------")
    print("|        Version string            |Major|Minor|Patch"
          "|unused|Endian|BP version|BP minor|Active|ColumnMajor|unused|")
    print("|          32 bytes                |  1B |  1B |  1B "
          "|  1B  |  1B  |    1B    |   1B   |  1B  |    1B     |  23B |")
    print("+----------------------------------+-----+-----+-----"
          "+------+------+----------+--------+------+-----------+------+")
    print("| {0} | {1} | {2} | {3} |      | {4}  "
          "|    {5}   |  {6}   | {7}  | {8}     |      |".format(
              versionStr, str(major).center(3), str(minor).center(3),
              str(micro).center(3), endian, str(bpversion).center(3),
              str(bpminorversion).center(3), activeStr, clmnStr))
    print("---------------------------------------------------------"
          "---------------------------------------------------------")
    return status


if __name__ == "__main__":
    print("ERROR: Utility main program is bp5dbg.py")



import numpy as np
import sys
import errno
import argparse
from adios2 import Stream
import itertools

verbose = False

def DetectMismatch(val1, val2, tolerance, tolperc):
    """
    Detects the mismatch between two scalar of any comparable type.

    Args:
        val1 : Any type scalar
        val2 : Any type scalar
        tolerance (float) : Tolerance used by the comparison
        tolperc (float) : Accepted tolerance between the two values as percentage

    Returns:
        bool: true (if there is a mismatch), false otherwise
    """
    mismatch = True
    if val1 == val2:
        return False
    if tolerance > 0:
        mismatch = (np.abs(val1 - val2) > tolerance)
    if tolperc > 0:
        if val1 == 0 and val2 != 0:
            mismatch = True
        else:
            mismatch = (np.abs(val1 - val2) / val1 > tolperc)
    if tolerance == 0 and tolperc == 0:
        mismatch = (val1 != val2)
    return mismatch

def CompareValues(array1, array2, tolerance, tolperc):
    """
    Detects the mismatch between two N dim arrays

    Args:
        array1 : N dimensional array
        array2 : N dimensional array
        tolerance (float) : Tolerance used by the comparison
        tolperc (float) : Accepted tolerance between the two values as percentage

    Returns:
        list(tuple): a list of mismatch values in the array: (index, value1, value2)
    """
    mismatch = []
    dimensions = array1.shape
    for idx in itertools.product(*[range(s) for s in dimensions]):
        if DetectMismatch(array1[idx], array2[idx], tolerance, tolperc):
            mismatch.append((idx, array1[idx], array2[idx]))
    return mismatch

def CompareFiles(filename1, filename2, tolerance, tolperc, stats=False):
    """
    Detects the mismatch between two BP files

    Args:
        filename1 : BP file name
        filename2 : BP file name
        tolerance (float) : Tolerance used by the comparison
        tolperc (float) : Accepted tolerance between the two values as percentage
        stats (bool) : boolean that constrols if stats are compared or not

    Returns:
        int: total number of mismatches between the two files
    """
    err = 0
    with Stream(filename1, 'rra') as s1, Stream(filename2, 'rra') as s2:
        # compare total number of steps
        num_steps = s1.num_steps()
        if num_steps != s2.num_steps():
            print(f"DIFFER : NUM_STEPS : {num_steps} <> {s2.num_steps()} : {filename1}"
                  f" : {filename2}")
            num_steps = min(num_steps, s2.num_steps())
            err += 1

        varInfoS1 = s1.available_variables()
        varInfoS2 = s2.available_variables()
        notinS1 = np.setdiff1d(list(varInfoS2.keys()), list(varInfoS1.keys()))
        notinS2 = np.setdiff1d(list(varInfoS1.keys()), list(varInfoS2.keys()))
        # compare total number of written variables
        if len(notinS1) > 0 or len(notinS1) > 0:
            print(f"DIFFER : AVAIL_VARS : {notinS1} <> {notinS2} : {filename1} : {filename2}")
            err += 1

        # compare each common variable
        for varName in np.intersect1d(list(varInfoS1.keys()), list(varInfoS2.keys())):
            # compare number of steps
            varSteps = int(varInfoS1[varName]["AvailableStepsCount"])
            if varSteps != int(varInfoS2[varName]["AvailableStepsCount"]):
                print(f"DIFFER : VARIABLE : {varName} : NUM_STEPS: "
                      f"{varInfoS1[varName]['AvailableStepsCount']} <> "
                      f"{varInfoS2[varName]['AvailableStepsCount']} : {filename1} : {filename2}")
                varSteps = min(varSteps, int(varInfoS2[varName]["AvailableStepsCount"]))
                err += 1

            # compare types
            if varInfoS1[varName]["Type"] != varInfoS2[varName]["Type"]:
                print(f"DIFFER : VARIABLE : {varName} : TYPE : {varInfoS1[varName]['Type']}"
                      f" <> {varInfoS2[varName]['Type']} : {filename1} : {filename2}")
                err += 1

            # compare dimensions (single value and shape)
            if varInfoS1[varName]["SingleValue"] != varInfoS2[varName]["SingleValue"]:
                print(f"DIFFER : VARIABLE : {varName} : SINGLE_VALUE : "
                      f"{varInfoS1[varName]['SingleValue']} <> "
                      f"{varInfoS2[varName]['SingleValue']} : {filename1} : {filename2}")
                err += 1
                # no point in comparing values if there is a difference in shape
                continue

            if varInfoS1[varName]["Shape"] != varInfoS2[varName]["Shape"]:
                print(f"DIFFER : VARIABLE : {varName} : SHAPE : {varInfoS1[varName]['Shape']}"
                      f" <> {varInfoS2[varName]['Shape']} : {filename1} : {filename2}")
                err += 1
                # no point in comparing values if there is a difference in shape
                continue

            if stats:
                # compare min/max
                if DetectMismatch(float(varInfoS1[varName]["Min"]),
                                  float(varInfoS2[varName]["Min"]),
                                  tolerance, tolperc):
                    print(f"DIFFER : VARIABLE : {varName} : MIN : {varInfoS1[varName]['Min']}"
                          f" <> {varInfoS2[varName]['Min']} : {filename1} : {filename2}")
                    err += 1
                if DetectMismatch(float(varInfoS1[varName]["Max"]),
                                  float(varInfoS2[varName]["Max"]),
                                  tolerance, tolperc):
                    print(f"DIFFER : VARIABLE : {varName} : MAX : {varInfoS1[varName]['Max']}"
                          f" <> {varInfoS2[varName]['Max']} : {filename1} : {filename2}")
                    err += 1
            # compare values
            for step in range(varSteps):
                values1 = s1.read(varName, step_selection=[step, 1])
                values2 = s2.read(varName, step_selection=[step, 1])
                ret = CompareValues(values1, values2, tolerance, tolperc)
                if len(ret) == 0:
                    continue
                err += len(ret)
                if verbose:
                    for (idx, value1, value2) in ret:
                        print(f"DIFFER : VARIABLE : {varName} : STEP : {step} : POSITION : "
                              f"{idx} : VALUE : {value1} <> {value2} : {filename1} : {filename2}")
                else:
                    print(f"DIFFER : VARIABLE : {varName} : STEP : {step} : VALUES_SUMMARY : "
                          f"{len(ret)} : {filename1} : {filename2}")

    return err


if __name__ == '__main__':
    description = 'Utility code to compare the data within BP files'
    epilog = 'If more than two files are provided, all files are compared to the first'
    parser = argparse.ArgumentParser(prog="bpcmp", description=description, epilog=epilog)

    parser.add_argument('bpfiles', metavar='file.bp', nargs='+',
                        help='list of BP files to be compared')
    parser.add_argument('-t', '--tolerance', dest='tolerance', type=float, default=0,
                        help='tolerance for comparing values (default: 0)')
    parser.add_argument('-p', '--perc-tolerance', dest='tolperc', type=float, default=0,
                        help='tolerance as percentage of values (default: 0)')
    parser.add_argument('-s', '--stats', action='store_true',
                        help='compare stats (min/max) and not values')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help='all different values are printed, not just a summary')

    args = parser.parse_args()
    if len(args.bpfiles) < 2:
        print("WARNING : At least two files need to be provided for any comparison")
        exit(errno.ENOEXEC)

    verbose = args.verbose

    diff_found = 0
    for i in range(1, len(args.bpfiles)):
        print(f"COMPARE : FILE : {args.bpfiles[0]} : FILE : {args.bpfiles[i]}")
        diff_file = CompareFiles(args.bpfiles[0], args.bpfiles[i], args.tolerance, args.tolperc,
                                 args.stats)
        if diff_file > 0:
            print(f"TOTAL_DIFFERENCES : {diff_file} : FILE : {args.bpfiles[0]} : FILE : "
                  f"{args.bpfiles[i]}")
            diff_found += 1

    if diff_found == 0:
        print("DONE : Files have the same content")



"""The ADIOS2 high-level API module
License:
  Distributed under the OSI-approved Apache License, Version 2.0.  See
  accompanying file Copyright.txt for details.
"""

import adios2.bindings

from adios2.adios import *
from adios2.attribute import *
from adios2.engine import *
from adios2.io import *
from adios2.operator import *
from adios2.stream import *
from adios2.variable import *
from adios2.file_reader import *
from adios2.bindings import (
    LocalValueDim,
    Mode,
    ShapeID,
    StepMode,
    StepStatus,
    DerivedVarType,
    Accuracy,
)

__license__ = "Apache-2.0"
__version__ = adios2.bindings.__version__
is_built_with_mpi = adios2.bindings.is_built_with_mpi

# JoinedDim = 2**64 - 2
JoinedDim = 18446744073709551614
